// CoreCalc, a spreadsheet core implementation 

// ----------------------------------------------------------------------
// Copyright (c) 2006 Thomas Iversen

// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without
// restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

//  * The above copyright notice and this permission notice shall be
//    included in all copies or substantial portions of the Software.

//  * The software is provided "as is", without warranty of any kind,
//    express or implied, including but not limited to the warranties of
//    merchantability, fitness for a particular purpose and
//    noninfringement.  In no event shall the authors or copyright
//    holders be liable for any claim, damages or other liability,
//    whether in an action of contract, tort or otherwise, arising from,
//    out of or in connection with the software or the use or other
//    dealings in the software.
// ----------------------------------------------------------------------

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Xml;
using System.Text.RegularExpressions;
using System.Xml.XPath;

// BenchmarkToPloticus
//
// Takes one or two TinyBench XML files and a Ploticus
// template.  If two XML files are given as argument
// they are merged into one dataset. The data is then
// substituted into a ploticus template file using 
// keywords (see generatePloticusPlotFile method).

namespace BenchmarkToPloticus {
  public enum MergeOrder { InbetweenFirstFit, SortedDescending, SortedAscending }

  public class Subdata {
    String name;
    String units;
    Double value;

    public Subdata(String name, String units, Double value) {
      this.units = units;
      this.name = name;
      this.value = value;
    }

    public Double Value {
      get { return value; }
    }

    public override string ToString() {
      return value.ToString();
    }
  }

  public class Data {
    String name;
    String color;
    List<Subdata> data;
    public Data(String name) {
      this.name = name;
      data = new List<Subdata>();
    }

    public void Add(Subdata subdata) {
      data.Add(subdata);
    }

    public String Color {
      get { return color; }
      set { color = value; }
    }

    public override String ToString() {
      StringBuilder sb = new StringBuilder();
      sb.Append("\"" + name + "\"");
      sb.Append("\t");
      foreach (Subdata sd in data) {
        sb.Append(sd.ToString());
      }
      return sb.ToString();
    }

    public long total() {
      long total = 0;
      foreach (Subdata sd in data) {
        total += (long)sd.Value;
      }
      return total;
    }

  }

  class Dataset {
    String name;
    String metaname;
    String units;
    String color;
    List<Data> dataset;

    public Dataset(String metaname, String name, String units, String color) {
      this.name = name;
      this.metaname = metaname;
      this.units = units;
      this.color = color;
      dataset = new List<Data>();
    }

    public String Color {
      get { return color; }
      set { color = value; }
    }

    private void MergeOne(Data data1) {
      long total = data1.total();
      int index = 0;
      long lasttotal = long.MaxValue;
      foreach (Data data in dataset) {
        long total1 = data.total();
        if ((lasttotal <= total && total <= total1) ||
            (total1 <= total && total <= lasttotal)) {
          dataset.Insert(index, data1);
          return;
        } else {
          lasttotal = total1;
          index++;
        }
      }

      // If we end up here it is possible that the merged data is the fastest

      if (total <= lasttotal)
        dataset.Add(data1);
    }

    private void MergeInbetweenFirstFit(List<Data> datalist) {
      // FIXME needs to be written more general

      foreach (Data data1 in datalist) {
        MergeOne(data1);
      }
    }

    public void Merge(Dataset dataset, MergeOrder mergeorder) {
      switch (mergeorder) {
        default:
        case MergeOrder.InbetweenFirstFit:
          MergeInbetweenFirstFit(dataset.dataset);
          break;
      }

    }

    public void Add(Data data) {
      data.Color = color;
      dataset.Add(data);
    }

    public String Name {
      get { return name; }
    }

    public String metaName {
      get { return metaname; }
    }


    public String Units {
      get { return units; }
    }

    private long xMax() {
      long max = 0;
      foreach (Data data in dataset) {
        max = Math.Max(max, data.total());
      }
      return max;
    }

    public String ColorList() {
      StringBuilder sb = new StringBuilder();
      sb.Append("colorlist: ");
      int index = 1;
      Boolean seenspecialcolors = false;
      foreach (Data data in dataset) {
        if (data.Color != null) {
          sb.Append(index);
          sb.Append(" ");
          sb.Append(data.Color);
          sb.Append(" ");
          seenspecialcolors = true;
        }
        index++;
      }
      if (seenspecialcolors) 
        return sb.ToString();
      else
        return "";
    }

    public String Xrange {
      get {
        long max = xMax();
        // round to nearest 100 above
        double ceil = Math.Ceiling((double)max / 100);
        long xmax = (long)ceil * 100;
        long xmin = 0;
        return xmin + " " + xmax.ToString();
      }
    }

    public String Yrange {
      get {
        long ymin = 0;
        long ymax = dataset.Count + 1;
        return ymin + " " + ymax;
      }
    }

    public override String ToString() {
      StringBuilder sb = new StringBuilder();
      foreach (Data data in dataset) {
        sb.Append(data.ToString());
        sb.AppendLine();
      }
      return sb.ToString();
    }
  }

  class Ploticus {
    String templateAsString;

    public void LoadTemplate(String templateFilename) {
      // Read ploticus template
      StreamReader sr = new StreamReader(templateFilename);
      templateAsString = sr.ReadToEnd();
      sr.Close();
    }

    public Ploticus(String templateFilename) {
      LoadTemplate(templateFilename);
    }

    public String generatePloticusPlotFile(Dataset dataset) {
      String s = Regex.Replace(templateAsString, "<TITLE>", dataset.metaName);
      s = Regex.Replace(s, "<DATA>", dataset.ToString());
      s = Regex.Replace(s, "<XRANGE>", dataset.Xrange);
      s = Regex.Replace(s, "<YRANGE>", dataset.Yrange);
      s = Regex.Replace(s, "<COLORLIST>", dataset.ColorList());
      s = Regex.Replace(s, "<XLABEL>", dataset.Name + " (" + dataset.Units + ")");
      return s;
    }
  }

  class Program {
    static void Usage() {
      Console.WriteLine("Usage: BenchmarkToPloticus <bench1.xml> "
                        + "<inpltemplate> <outplscript> [<bench2.xml>]");
    }

    static void Main(string[] args) {
      if (args != null && args.Length != 3 && args.Length != 4) {
        Usage();
      } else {
        String infile = args[0];
        String template = args[1];
        String outfile = args[2];

        Dataset dataset = LoadBenchmarkData(infile, null);

        Dataset dataset1 = null;
        if (args.Length == 4) {
          dataset1 = LoadBenchmarkData(args[3], "yellow");
          dataset.Merge(dataset1, MergeOrder.InbetweenFirstFit);
        }

        Ploticus pl = new Ploticus(template);
        String ploticusplotfile = pl.generatePloticusPlotFile(dataset);

        StreamWriter sw = new StreamWriter(outfile);
        sw.Write(ploticusplotfile);
        sw.Close();
      }
    }

    static Dataset LoadBenchmarkData(String filename, String color) {
      XPathDocument doc = new XPathDocument(filename);
      XPathNavigator nav = doc.CreateNavigator();

      XPathExpression metaexpr = nav.Compile("/Root/MetaData/Title");
      XPathNodeIterator metaiterator = nav.Select(metaexpr);

      String metatitle = null;
      while (metaiterator.MoveNext()) {
        XPathNavigator metanav = metaiterator.Current.Clone();
        metatitle = metanav.Value;
      }

      XPathExpression dsexpr = nav.Compile("/Root/DataSet");
      XPathNodeIterator dsiterator = nav.Select(dsexpr);

      String datasettitle = null;
      String datasetunits = null;
      while (dsiterator.MoveNext()) {
        XPathNavigator dsnav = dsiterator.Current.Clone();
        datasettitle = dsnav.GetAttribute("Name", "");
        datasetunits = dsnav.GetAttribute("Units", "");
      }

      Dataset dataset = new Dataset(metatitle, datasettitle, datasetunits, color);

      XPathExpression expr = nav.Compile("/Root/DataSet/Data");
      XPathNodeIterator iterator = nav.Select(expr);
      while (iterator.MoveNext()) {
        XPathNavigator nav2 = iterator.Current.Clone();
        String datatitle = nav2.GetAttribute("Name", "");

        Data data = new Data(datatitle);
        dataset.Add(data);

        XPathExpression expr1 = nav2.Compile("SubData");
        XPathNodeIterator iterator1 = nav2.Select(expr1);
        while (iterator1.MoveNext()) {
          XPathNavigator nav3 = iterator1.Current.Clone();
          int noruns = int.Parse(nav3.GetAttribute("NumberOfRuns", ""));
          String units = nav3.GetAttribute("Units", "");
          String name = nav3.GetAttribute("Name", "");
          XPathExpression expr2 = nav3.Compile("Run");
          XPathNodeIterator iterator2 = nav3.Select(expr2);

          double total = 0;
          while (iterator2.MoveNext()) {
            XPathNavigator nav4 = iterator2.Current.Clone();
            double value;
            if (double.TryParse(nav4.Value, out value))
              total += value;
          }
          Subdata subdata = new Subdata(name, units, Math.Round(total / noruns));
          data.Add(subdata);
        }
      }
      return dataset;
    }
  }
}
