// CoreCalc, a spreadsheet core implementation 

// ----------------------------------------------------------------------
// Copyright (c) 2006 Peter Sestoft

// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without
// restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

//  * The above copyright notice and this permission notice shall be
//    included in all copies or substantial portions of the Software.

//  * The software is provided "as is", without warranty of any kind,
//    express or implied, including but not limited to the warranties of
//    merchantability, fitness for a particular purpose and
//    noninfringement.  In no event shall the authors or copyright
//    holders be liable for any claim, damages or other liability,
//    whether in an action of contract, tort or otherwise, arising from,
//    out of or in connection with the software or the use or other
//    dealings in the software.
// ----------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;                        // MemoryStream, Stream
using System.Text;

// The Cell class and its subclasses represent the possible contents 
// of a spreadsheet cell.

namespace CoreCalc {

  // A Cell is what populates one cell of a Sheet (if anything)

  public abstract class Cell {
    public abstract Value Eval(Sheet sheet, int col, int row);

    public abstract void Apply(Sheet sheet, int col, int row, Act<Value> act);

    public abstract Cell MoveContents(int deltaCol, int deltaRow);

    public abstract void InsertRowCols(Dictionary<Expr, Adjusted<Expr>> adjusted,
               Sheet modSheet, bool thisSheet,
               int R, int N, int r, bool doRows);

    public abstract void Reset();

    // Show computed value
    public String ShowValue(Sheet sheet, int col, int row) {
      Value v = Eval(sheet, col, row);
      return v != null ? v.ToString() : "";
    }

    // Show constant or formula or matrix formula
    public abstract String Show(int col, int row, Formats fo);

    // Parse string to cell contents at (col, row) in given workbook 
    public static Cell Parse(String text, Workbook workbook, int col, int row) {
      Scanner scanner = new Scanner(MakeStream(text));
      Parser parser = new Parser(scanner);
      Cell cell = parser.ParseCell(workbook, col, row);
      return cell;
    }

    private static Stream MakeStream(String s) {
      char[] cs = s.ToCharArray();
      byte[] bs = new byte[cs.Length];
      for (int i = 0; i < cs.Length; i++)
        bs[i] = (byte)(cs[i]);
      return new MemoryStream(bs);
    }
  }

  // ----------------------------------------------------------------
  // A ConstCell is a cell that contains a constant only.  
  // In contrast to general cells, it is immutable and can be shared

  abstract class ConstCell : Cell {
    public override Cell MoveContents(int deltaCol, int deltaRow) {
      return this;
    }

    public override void InsertRowCols(Dictionary<Expr, Adjusted<Expr>> adjusted,
               Sheet modSheet, bool thisSheet,
               int R, int N, int r, bool doRows) { }

    public override void Reset() { }
  }

  // ----------------------------------------------------------------
  // A NumberCell is a cell containing a floating-point constant;
  // not a constant number-valued expression

  sealed class NumberCell : ConstCell {
    public readonly NumberValue value;        // Non-null

    public NumberCell(double d) {
      this.value = new NumberValue(d);
    }

    public override Value Eval(Sheet sheet, int col, int row) {
      return value;
    }

    public override void Apply(Sheet sheet, int col, int row, Act<Value> act) {
      act(value);
    }

    public override String Show(int col, int row, Formats fo) {
      return value.value.ToString();
    }
  }

  // ----------------------------------------------------------------
  // A TextCell is a call containing a string constant; 
  // not a constant string-valued expression

  sealed class TextCell : ConstCell {
    public readonly TextValue value;          // Non-null

    public TextCell(String s) {
      this.value = new TextValue(s);
    }

    public override Value Eval(Sheet sheet, int col, int row) {
      return value;
    }

    public override void Apply(Sheet sheet, int col, int row, Act<Value> act) {
      act(value);
    }

    public override String Show(int col, int row, Formats fo) {
      return "'" + value.value;
    }
  }

  // ----------------------------------------------------------------
  // A Formula is a non-null caching expression contained in one cell.

  sealed class Formula : Cell {
    public readonly Workbook workbook;  // Non-null
    private Expr e;                     // Non-null
    private bool visited, uptodate;
    private Value v;                  // Up to date if uptodate==workbook.Set

    public Formula(Workbook workbook, Expr e) {
      if (workbook == null || e == null)
        throw new Exception("Formula arguments");
      else {
        this.workbook = workbook;
        this.e = e;
        this.visited = this.uptodate = workbook.Set;
      }
    }

    public Formula(Formula f) : this(f.workbook, f.e) { }

    // FIXME: Adequate for moving one cell, but block moves and row/column
    // inserts should avoid the duplication and unsharing of expressions. 
    public override Cell MoveContents(int deltaCol, int deltaRow) {
      return new Formula(workbook, e.Move(deltaCol, deltaRow));
    }

    // Evaluate expression if necessary, and cache its value
    public override Value Eval(Sheet sheet, int col, int row) {
      if (uptodate != workbook.Set) {
        if (visited == workbook.Set)
          throw new CyclicException("Cyclic cell reference: "
                                    + Show(col, row, workbook.Format));
        else {
          visited = workbook.Set;
          try {
            v = e.Eval(sheet, col, row);
            uptodate = workbook.Set;
          } catch (ErrorValueException exn) {
            v = exn.errorValue;
          }
        }
      }
      return v;
    }

    public override void Apply(Sheet sheet, int col, int row, Act<Value> act) {
      act(Eval(sheet, col, row));
    }

    public override void InsertRowCols(Dictionary<Expr, Adjusted<Expr>> adjusted,
               Sheet modSheet, bool thisSheet,
               int R, int N, int r, bool doRows) {
      Adjusted<Expr> ae;
      if (adjusted.ContainsKey(e) && r < adjusted[e].upper)
        // There is a valid cached adjusted expression
        ae = adjusted[e];
      else {
        // Compute a new adjusted expression and insert into the cache
        ae = e.InsertRowCols(modSheet, thisSheet, R, N, r, doRows);
        Console.WriteLine("Making new adjusted at rowcol " + r
                          + "; upper = " + ae.upper);
        if (ae.same) { // For better sharing, reuse unadjusted e if same
          ae = new Adjusted<Expr>(e, ae.upper, ae.same);
          Console.WriteLine("Reusing expression");
        }
        adjusted[e] = ae;
      }
      Debug.Assert(r < ae.upper, "Formula.InsertRowCols");
      e = ae.e;
    }

    // Reset recomputation flags after a circularity has been found
    public override void Reset() {
      visited = uptodate = workbook.Set;
    }

    public override String Show(int col, int row, Formats fo) {
      return "=" + e.Show(col, row, 0, fo);
    }

    public Expr Expr {
      get { return e; }
      set { e = value; }
    }
  }

  // ----------------------------------------------------------------
  // A MatrixFormula is a cached matrix formula shared among several
  // cells, each cell accessing one part of the result matrix.  Several
  // MatrixFormula cells share one CachedMatrixFormula cell; evaluation
  // of one cell will evaluate the formula and cache its (matrix) value
  // for the other cells to use.

  sealed class MatrixFormula : Cell {
    private readonly CachedMatrixFormula cmf;  // Non-null
    private readonly CellAddr ca;  // Cell's location within matrix value

    public MatrixFormula(CachedMatrixFormula cmf, CellAddr ca) {
      this.cmf = cmf; this.ca = ca;
    }

    public MatrixFormula(CachedMatrixFormula cmf, int col, int row)
      : this(cmf, new CellAddr(col, row)) { }

    public override Value Eval(Sheet sheet, int col, int row) {
      MatrixValue matrix = cmf.Eval(sheet, col, row);
      if (matrix != null)
        return matrix[ca];
      else
        return new ErrorValue("NONMATRIX");
    }

    public override void Apply(Sheet sheet, int col, int row, Act<Value> act) {
      cmf.Apply(sheet, col, row, act);
    }

    public bool Contains(int col, int row) {
      return cmf.ulCa.col <= col && col <= cmf.lrCa.col
          && cmf.ulCa.row <= row && row <= cmf.lrCa.row;
    }

    public override Cell MoveContents(int deltaCol, int deltaRow) {
      // FIXME: loses sharing of the CachedMatrixFormula; but then again
      // a matrix formula should never be moved cell by cell, but in its
      // entirety, and in one go.
      return new MatrixFormula(cmf.MoveContents(deltaCol, deltaRow), ca);
    }

    public override void InsertRowCols(Dictionary<Expr, Adjusted<Expr>> adjusted,
               Sheet modSheet, bool thisSheet,
               int R, int N, int r, bool doRows) {
      // FIXME: Update underlying formula only once!
    }

    public override void Reset() {
      cmf.Reset();
    }

    public override String Show(int col, int row, Formats fo) {
      return "{" + cmf.Show(col, row, fo) + "}";
    }
  }

  // ----------------------------------------------------------------
  // A cached matrix formula is shared between multiple matrix cells.
  // It contains a (caching) matrix-valued formula, the original cell
  // address of that formula, and the upper left and lower right corners
  // of the matrix.

  sealed class CachedMatrixFormula {
    public readonly Formula formula;             // Non-null
    private readonly int formulaCol, formulaRow; // Location of formula entry
    public readonly CellAddr ulCa, lrCa;         // Corners of matrix formula
    // Invariant: Every cell within sheet[ulCa, lrCa] is a MatrixFormula
    // whose CachedMatrixFormula instance is this one.

    public CachedMatrixFormula(Formula formula, int formulaCol, int formulaRow,
                               CellAddr ulCa, CellAddr lrCa) {
      if (formula == null)
        throw new Exception("CachedMatrixFormula arguments");
      else {
        this.formula = formula;
        this.formulaCol = formulaCol;
        this.formulaRow = formulaRow;
        this.ulCa = ulCa;
        this.lrCa = lrCa;
      }
    }

    // Evaluate expression if necessary, cast, and return
    public MatrixValue Eval(Sheet sheet, int col, int row) {
      return formula.Eval(sheet, formulaCol, formulaRow) as MatrixValue;
    }

    public void Apply(Sheet sheet, int col, int row, Act<Value> act) {
      formula.Apply(sheet, formulaCol, formulaRow, act);
    }

    public CachedMatrixFormula MoveContents(int deltaCol, int deltaRow) {
      // FIXME: Unshares the formula, shouldn't ...
      return new CachedMatrixFormula((Formula)formula.MoveContents(deltaCol, deltaRow),
                                     formulaCol, formulaRow, ulCa, lrCa);
    }

    // Reset recomputation flags after a circularity has been found
    public void Reset() {
      formula.Reset();
    }

    public String Show(int col, int row, Formats fo) {
      return formula.Show(col, row, fo);
    }
  }
}
