
using System;
using System.IO;
using System.Collections;

namespace CoreCalc {

public class Token {
	public int kind;    // token kind
	public int pos;     // token position in the source text (starting at 0)
	public int col;     // token column (starting at 0)
	public int line;    // token line (starting at 1)
	public string val;  // token value
	public Token next;  // ML 2005-03-11 Tokens are kept in linked list
}

public class Buffer {
	public const char EOF = (char)256;
	const int MAX_BUFFER_LENGTH = 64 * 1024; // 64KB
	byte[] buf;         // input buffer
	int bufStart;       // position of first byte in buffer relative to input stream
	int bufLen;         // length of buffer
	int fileLen;        // length of input stream
	int pos;            // current position in buffer
	Stream stream;      // input stream (seekable)
	bool isUserStream;  // was the stream opened by the user?
	
	public Buffer (Stream s, bool isUserStream) {
		stream = s; this.isUserStream = isUserStream;
		fileLen = bufLen = (int) s.Length;
		if (stream.CanSeek && bufLen > MAX_BUFFER_LENGTH) bufLen = MAX_BUFFER_LENGTH;
		buf = new byte[bufLen];
		bufStart = Int32.MaxValue; // nothing in the buffer so far
		Pos = 0; // setup buffer to position 0 (start)
		if (bufLen == fileLen) Close();
	}
	
	~Buffer() { Close(); }
	
	void Close() {
		if (!isUserStream && stream != null) {
			stream.Close();
			stream = null;
		}
	}
	
	public int Read () {
		if (pos < bufLen) {
			return buf[pos++];
		} else if (Pos < fileLen) {
			Pos = Pos; // shift buffer start to Pos
			return buf[pos++];
		} else {
			return EOF;
		}
	}

	public int Peek () {
		if (pos < bufLen) {
			return buf[pos];
		} else if (Pos < fileLen) {
			Pos = Pos; // shift buffer start to Pos
			return buf[pos];
		} else {
			return EOF;
		}
	}
	
	public string GetString (int beg, int end) {
		int len = end - beg;
		char[] buf = new char[len];
		int oldPos = Pos;
		Pos = beg;
		for (int i = 0; i < len; i++) buf[i] = (char) Read();
		Pos = oldPos;
		return new String(buf);
	}

	public int Pos {
		get { return pos + bufStart; }
		set {
			if (value < 0) value = 0; 
			else if (value > fileLen) value = fileLen;
			if (value >= bufStart && value < bufStart + bufLen) { // already in buffer
				pos = value - bufStart;
			} else if (stream != null) { // must be swapped in
				stream.Seek(value, SeekOrigin.Begin);
				bufLen = stream.Read(buf, 0, buf.Length);
				bufStart = value; pos = 0;
			} else {
				pos = fileLen - bufStart; // make Pos return fileLen
			}
		}
	}
}

public class Scanner {
	const char EOL = '\n';
	const int eofSym = 0; /* pdt */
	const int charSetSize = 256;
	const int maxT = 34;
	const int noSym = 34;
	short[] start = {
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0, 26,  0,  9,  0, 48, 28, 54, 55, 59, 46, 58, 47,  0, 60,
	 30, 30, 30, 30, 30, 30, 30, 30, 30, 30, 53, 57, 62, 61, 63,  0,
	  0, 31, 31, 31, 31, 31, 31, 31, 31, 31, 29, 29, 29, 29, 29, 29,
	 29, 29, 32, 29, 29, 29, 29, 29, 29, 29, 29,  0,  0,  0, 56,  0,
	  0, 31, 31, 31, 31, 31, 31, 31, 31, 31, 29, 29, 29, 29, 29, 29,
	 29, 29, 29, 29, 29, 29, 29, 29, 29, 29, 29,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	  -1};


	public Buffer buffer; // scanner buffer
	
	Token t;          // current token
	char ch;          // current input character
	int pos;          // column number of current character
	int line;         // line number of current character
	int lineStart;    // start position of current line
	int oldEols;      // EOLs that appeared in a comment;
	BitArray ignore;  // set of characters to be ignored by the scanner

	Token tokens;     // list of tokens already peeked (first token is a dummy)
	Token pt;         // current peek token
	
	char[] tval = new char[128]; // text of current token
	int tlen;         // length of current token
	
	public Scanner (string fileName) {
		try {
			Stream stream = new FileStream(fileName, FileMode.Open, FileAccess.Read, FileShare.Read);
			buffer = new Buffer(stream, false);
			Init();
		} catch (IOException) {
			Console.WriteLine("--- Cannot open file {0}", fileName);
			System.Environment.Exit(1);
		}
	}
	
	public Scanner (Stream s) {
		buffer = new Buffer(s, true);
		Init();
	}
	
	void Init() {
		pos = -1; line = 1; lineStart = 0;
		oldEols = 0;
		NextCh();
		ignore = new BitArray(charSetSize+1);
		ignore[' '] = true;  // blanks are always white space
		ignore[9] = true; ignore[10] = true; ignore[13] = true; 
		pt = tokens = new Token();  // first token is a dummy
	}
	
	void NextCh() {
		if (oldEols > 0) { ch = EOL; oldEols--; } 
		else {
			ch = (char)buffer.Read(); pos++;
			// replace isolated '\r' by '\n' in order to make
			// eol handling uniform across Windows, Unix and Mac
			if (ch == '\r' && buffer.Peek() != '\n') ch = EOL;
			if (ch == EOL) { line++; lineStart = pos + 1; }
		}

	}

	void AddCh() {
		if (tlen >= tval.Length) {
			char[] newBuf = new char[2 * tval.Length];
			Array.Copy(tval, 0, newBuf, 0, tval.Length);
			tval = newBuf;
		}
		tval[tlen++] = ch;
		NextCh();
	}



	bool Comment0() {
		int level = 1, line0 = line, lineStart0 = lineStart;
		NextCh();
		if (ch == '/') {
			NextCh();
			for(;;) {
				if (ch == 13) {
					NextCh();
					if (ch == 10) {
						level--;
						if (level == 0) { oldEols = line - line0; NextCh(); return true; }
						NextCh();
					}
				} else if (ch == Buffer.EOF) return false;
				else NextCh();
			}
		} else {
			if (ch==EOL) {line--; lineStart = lineStart0;}
			pos = pos - 2; buffer.Pos = pos+1; NextCh();
		}
		return false;
	}

	bool Comment1() {
		int level = 1, line0 = line, lineStart0 = lineStart;
		NextCh();
		if (ch == '*') {
			NextCh();
			for(;;) {
				if (ch == '*') {
					NextCh();
					if (ch == '/') {
						level--;
						if (level == 0) { oldEols = line - line0; NextCh(); return true; }
						NextCh();
					}
				} else if (ch == '/') {
					NextCh();
					if (ch == '*') {
						level++; NextCh();
					}
				} else if (ch == Buffer.EOF) return false;
				else NextCh();
			}
		} else {
			if (ch==EOL) {line--; lineStart = lineStart0;}
			pos = pos - 2; buffer.Pos = pos+1; NextCh();
		}
		return false;
	}


	void CheckLiteral() {
		switch (t.val) {
			default: break;
		}
	}

	Token NextToken() {
		while (ignore[ch]) NextCh();
		if (ch == '/' && Comment0() ||ch == '/' && Comment1()) return NextToken();
		int apx = 0;
		t = new Token();
		t.pos = pos; t.col = pos - lineStart + 1; t.line = line; 
		int state = start[ch];
		tlen = 0; AddCh();
		
		switch (state) {
			case -1: { t.kind = eofSym; break; } // NextCh already done
			case 0: { t.kind = noSym; break; }   // NextCh already done
			case 1:
				{
					tlen -= apx;
					pos = pos - apx - 1; line = t.line;
					buffer.Pos = pos + 1; NextCh();
					t.kind = 1; break;}
			case 2:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 4;}
				else if ((ch == '+' || ch == '-')) {AddCh(); goto case 3;}
				else {t.kind = noSym; break;}
			case 3:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 4;}
				else {t.kind = noSym; break;}
			case 4:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 4;}
				else {t.kind = 2; break;}
			case 5:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 6;}
				else {t.kind = noSym; break;}
			case 6:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 6;}
				else if ((ch == 'E' || ch == 'e')) {AddCh(); goto case 2;}
				else {t.kind = 2; break;}
			case 7:
				if ((ch == '!')) {AddCh(); goto case 8;}
				else if ((ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'Z' || ch >= 'a' && ch <= 'z')) {AddCh(); goto case 7;}
				else {t.kind = noSym; break;}
			case 8:
				{t.kind = 3; break;}
			case 9:
				if ((ch >= 'A' && ch <= 'I' || ch >= 'a' && ch <= 'i')) {AddCh(); goto case 10;}
				else {t.kind = noSym; break;}
			case 10:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 12;}
				else if ((ch == '$')) {AddCh(); goto case 11;}
				else if ((ch >= 'A' && ch <= 'Z' || ch >= 'a' && ch <= 'z')) {AddCh(); goto case 10;}
				else {t.kind = noSym; break;}
			case 11:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 12;}
				else {t.kind = noSym; break;}
			case 12:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 12;}
				else {t.kind = 4; break;}
			case 13:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 15;}
				else if ((ch == '+' || ch == '-')) {AddCh(); goto case 14;}
				else {t.kind = noSym; break;}
			case 14:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 15;}
				else {t.kind = noSym; break;}
			case 15:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 15;}
				else if (ch == ']') {AddCh(); goto case 16;}
				else {t.kind = noSym; break;}
			case 16:
				{t.kind = 7; break;}
			case 17:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 19;}
				else if ((ch == '+' || ch == '-')) {AddCh(); goto case 18;}
				else {t.kind = noSym; break;}
			case 18:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 19;}
				else {t.kind = noSym; break;}
			case 19:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 19;}
				else if (ch == ']') {AddCh(); goto case 20;}
				else {t.kind = noSym; break;}
			case 20:
				{t.kind = 10; break;}
			case 21:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 21;}
				else {t.kind = 12; break;}
			case 22:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 24;}
				else if ((ch == '+' || ch == '-')) {AddCh(); goto case 23;}
				else {t.kind = noSym; break;}
			case 23:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 24;}
				else {t.kind = noSym; break;}
			case 24:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 24;}
				else if (ch == ']') {AddCh(); goto case 25;}
				else {t.kind = noSym; break;}
			case 25:
				{t.kind = 13; break;}
			case 26:
				if ((ch <= 9 || ch >= 11 && ch <= 12 || ch >= 14 && ch <= '!' || ch >= '#' && ch <= '[' || ch >= ']' && ch <= 255)) {AddCh(); goto case 26;}
				else if (ch == '"') {AddCh(); goto case 27;}
				else {t.kind = noSym; break;}
			case 27:
				{t.kind = 14; break;}
			case 28:
				if ((ch <= 9 || ch >= 11 && ch <= 12 || ch >= 14 && ch <= '[' || ch >= ']' && ch <= 255)) {AddCh(); goto case 28;}
				else {t.kind = 15; break;}
			case 29:
				if ((ch >= 'A' && ch <= 'Z' || ch >= 'a' && ch <= 'z')) {AddCh(); goto case 29;}
				else if ((ch == '!')) {AddCh(); goto case 8;}
				else if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 7;}
				else if (ch == '(') {apx++; AddCh(); goto case 1;}
				else {t.kind = noSym; break;}
			case 30:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 30;}
				else if ((ch == '!')) {AddCh(); goto case 8;}
				else if ((ch >= 'A' && ch <= 'D' || ch >= 'F' && ch <= 'Z' || ch >= 'a' && ch <= 'd' || ch >= 'f' && ch <= 'z')) {AddCh(); goto case 7;}
				else if ((ch == 'E' || ch == 'e')) {AddCh(); goto case 33;}
				else if (ch == '.') {AddCh(); goto case 5;}
				else {t.kind = 2; break;}
			case 31:
				if ((ch >= 'A' && ch <= 'Z' || ch >= 'a' && ch <= 'z')) {AddCh(); goto case 31;}
				else if ((ch == '!')) {AddCh(); goto case 8;}
				else if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 34;}
				else if ((ch == '$')) {AddCh(); goto case 11;}
				else if (ch == '(') {apx++; AddCh(); goto case 1;}
				else {t.kind = noSym; break;}
			case 32:
				if ((ch >= 'A' && ch <= 'B' || ch >= 'D' && ch <= 'Z' || ch >= 'a' && ch <= 'z')) {AddCh(); goto case 29;}
				else if ((ch == '!')) {AddCh(); goto case 8;}
				else if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 35;}
				else if (ch == '(') {apx++; AddCh(); goto case 1;}
				else if (ch == 'C') {AddCh(); goto case 36;}
				else if (ch == '[') {AddCh(); goto case 37;}
				else {t.kind = noSym; break;}
			case 33:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 38;}
				else if ((ch == '!')) {AddCh(); goto case 8;}
				else if ((ch >= 'A' && ch <= 'Z' || ch >= 'a' && ch <= 'z')) {AddCh(); goto case 7;}
				else if ((ch == '+' || ch == '-')) {AddCh(); goto case 3;}
				else {t.kind = noSym; break;}
			case 34:
				if ((ch == '!')) {AddCh(); goto case 8;}
				else if ((ch >= 'A' && ch <= 'Z' || ch >= 'a' && ch <= 'z')) {AddCh(); goto case 7;}
				else if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 34;}
				else {t.kind = 4; break;}
			case 35:
				if ((ch == '!')) {AddCh(); goto case 8;}
				else if ((ch >= 'A' && ch <= 'B' || ch >= 'D' && ch <= 'Z' || ch >= 'a' && ch <= 'z')) {AddCh(); goto case 7;}
				else if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 35;}
				else if (ch == 'C') {AddCh(); goto case 39;}
				else {t.kind = noSym; break;}
			case 36:
				if ((ch >= 'A' && ch <= 'Z' || ch >= 'a' && ch <= 'z')) {AddCh(); goto case 29;}
				else if ((ch == '!')) {AddCh(); goto case 8;}
				else if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 40;}
				else if (ch == '(') {apx++; AddCh(); goto case 1;}
				else if (ch == '[') {AddCh(); goto case 13;}
				else {t.kind = 5; break;}
			case 37:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 41;}
				else if ((ch == '+' || ch == '-')) {AddCh(); goto case 42;}
				else {t.kind = noSym; break;}
			case 38:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 38;}
				else if ((ch == '!')) {AddCh(); goto case 8;}
				else if ((ch >= 'A' && ch <= 'Z' || ch >= 'a' && ch <= 'z')) {AddCh(); goto case 7;}
				else {t.kind = 2; break;}
			case 39:
				if ((ch == '!')) {AddCh(); goto case 8;}
				else if ((ch >= 'A' && ch <= 'Z' || ch >= 'a' && ch <= 'z')) {AddCh(); goto case 7;}
				else if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 43;}
				else if (ch == '[') {AddCh(); goto case 17;}
				else {t.kind = 8; break;}
			case 40:
				if ((ch == '!')) {AddCh(); goto case 8;}
				else if ((ch >= 'A' && ch <= 'Z' || ch >= 'a' && ch <= 'z')) {AddCh(); goto case 7;}
				else if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 40;}
				else {t.kind = 6; break;}
			case 41:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 41;}
				else if (ch == ']') {AddCh(); goto case 44;}
				else {t.kind = noSym; break;}
			case 42:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 41;}
				else {t.kind = noSym; break;}
			case 43:
				if ((ch == '!')) {AddCh(); goto case 8;}
				else if ((ch >= 'A' && ch <= 'Z' || ch >= 'a' && ch <= 'z')) {AddCh(); goto case 7;}
				else if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 43;}
				else {t.kind = 9; break;}
			case 44:
				if (ch == 'C') {AddCh(); goto case 45;}
				else {t.kind = noSym; break;}
			case 45:
				if ((ch >= '0' && ch <= '9')) {AddCh(); goto case 21;}
				else if (ch == '[') {AddCh(); goto case 22;}
				else {t.kind = 11; break;}
			case 46:
				{t.kind = 16; break;}
			case 47:
				{t.kind = 17; break;}
			case 48:
				{t.kind = 18; break;}
			case 49:
				{t.kind = 19; break;}
			case 50:
				{t.kind = 20; break;}
			case 51:
				{t.kind = 22; break;}
			case 52:
				{t.kind = 24; break;}
			case 53:
				{t.kind = 25; break;}
			case 54:
				{t.kind = 26; break;}
			case 55:
				{t.kind = 27; break;}
			case 56:
				{t.kind = 28; break;}
			case 57:
				{t.kind = 29; break;}
			case 58:
				{t.kind = 30; break;}
			case 59:
				{t.kind = 31; break;}
			case 60:
				{t.kind = 32; break;}
			case 61:
				if (ch == '=') {AddCh(); goto case 49;}
				else {t.kind = 33; break;}
			case 62:
				if (ch == '>') {AddCh(); goto case 50;}
				else if (ch == '=') {AddCh(); goto case 51;}
				else {t.kind = 21; break;}
			case 63:
				if (ch == '=') {AddCh(); goto case 52;}
				else {t.kind = 23; break;}

		}
		t.val = new String(tval, 0, tlen);
		return t;
	}
	
	// get the next token (possibly a token already seen during peeking)
	public Token Scan () {
		if (tokens.next == null) {
			return NextToken();
		} else {
			pt = tokens = tokens.next;
			return tokens;
		}
	}

	// peek for the next token, ignore pragmas
	public Token Peek () {
		if (pt.next == null) {
			do {
				pt = pt.next = NextToken();
			} while (pt.kind > maxT); // skip pragmas
		} else {
			do {
				pt = pt.next;
			} while (pt.kind > maxT);
		}
		return pt;
	}
	
	// make sure that peeking starts at the current scan position
	public void ResetPeek () { pt = tokens; }

} // end Scanner

}