// CoreCalc, a spreadsheet core implementation 

// ----------------------------------------------------------------------
// Copyright (c) 2006 Peter Sestoft

// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without
// restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

//  * The above copyright notice and this permission notice shall be
//    included in all copies or substantial portions of the Software.

//  * The software is provided "as is", without warranty of any kind,
//    express or implied, including but not limited to the warranties of
//    merchantability, fitness for a particular purpose and
//    noninfringement.  In no event shall the authors or copyright
//    holders be liable for any claim, damages or other liability,
//    whether in an action of contract, tort or otherwise, arising from,
//    out of or in connection with the software or the use or other
//    dealings in the software.
// ----------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Text;

// Class Expr and its subclasses are used to recursively build formulas

namespace CoreCalc {
  // ----------------------------------------------------------------
  // Expressions are used to build formulas

  public abstract class Expr {
    // Update cell references when containing cell is moved (not copied)
    public abstract Expr Move(int deltaCol, int deltaRow);

    // Evaluate expression as if at cell address sheet[col, row]
    public abstract Value Eval(Sheet sheet, int col, int row);

    // Apply act to result of evaluating expression at sheet[col, row]
    public abstract void Apply(Sheet sheet, int col, int row, Act<Value> act);

    // Insert N new rows before row R
    public abstract Adjusted<Expr> InsertRowCols(Sheet modSheet, bool thisSheet,
                   int R, int N, int r,
                   bool doRows);

    // Show contents as expression
    public abstract String Show(int col, int row, int ctxpre, Formats fo);
  }

  // ----------------------------------------------------------------
  // A Const expression is a constant (immutable, sharable)

  abstract class Const : Expr {
    public override Expr Move(int deltaCol, int deltaRow) {
      return this;
    }

    public override Adjusted<Expr> InsertRowCols(Sheet modSheet, bool thisSheet,
                                                 int R, int N, int r,
                                                 bool doRows) 
    {
      return new Adjusted<Expr>(this);
    }
  }

  // ----------------------------------------------------------------
  // A NumberConst is a constant expression

  class NumberConst : Const {
    private readonly NumberValue value;

    public NumberConst(double d) {
      value = new NumberValue(d);
    }

    public override Value Eval(Sheet sheet, int col, int row) {
      return value;
    }

    public override void Apply(Sheet sheet, int col, int row, Act<Value> act) {
      act(value);
    }

    public override String Show(int col, int row, int ctxpre, Formats fo) {
      return value.ToString();
    }
  }

  // ----------------------------------------------------------------
  // A TextConst is a constant string-valued expression

  class TextConst : Const {
    private readonly TextValue value;

    public TextConst(String s) {
      value = new TextValue(s);
    }

    public override Value Eval(Sheet sheet, int col, int row) {
      return value;
    }

    public override void Apply(Sheet sheet, int col, int row, Act<Value> act) {
      act(value);
    }

    public override String Show(int col, int row, int ctxpre, Formats fo) {
      return "\"" + value.ToString() + "\"";
    }
  }

  // ----------------------------------------------------------------
  // An Error expression represents a static error, e.g. invalid reference

  class Error : Expr {
    private readonly String error;
    private readonly ErrorValue value;

    public Error(String s) {
      value = new ErrorValue(s);
      error = "#" + s;
    }

    public override Value Eval(Sheet sheet, int col, int row) {
      return value;
    }

    public override void Apply(Sheet sheet, int col, int row, Act<Value> act) {
      act(value);
    }

    public override String Show(int col, int row, int ctxpre, Formats fo) {
      return error;
    }

    public override Expr Move(int deltaCol, int deltaRow) {
      return this;
    }

    public override Adjusted<Expr> InsertRowCols(Sheet modSheet, bool thisSheet,
                                                 int R, int N, int r,
                                                 bool doRows) 
    {
      return new Adjusted<Expr>(this);
    }
  }

  // ----------------------------------------------------------------
  // A FunCall is an operator application such as 1+$A$4 or a function
  // call such as RAND() or SIN(4*A$7) or SUM(B4:B52; 3) or IF(A1; A2; 1/A1)

  class FunCall : Expr {
    private readonly Function function;   // Non-null
    private readonly Expr[] es;           // Non-null, elements non-null

    public FunCall(Function function, Expr[] es) {
      if (function == null || es == null
          || Array.Exists(es, delegate(Expr e) { return e == null; }))
        throw new Exception("Function or argument null");
      else {
        this.function = function;
        this.es = es;
      }
    }

    public FunCall(String name, Expr[] es)
      : this(Function.Get(name), es) { }

    // Arguments are passed unevaluated to cater for non-strict IF 

    public override Value Eval(Sheet sheet, int col, int row) {
      return function.applier(sheet, es, col, row);
    }

    public override void Apply(Sheet sheet, int col, int row, Act<Value> act) {
      Value v = function.applier(sheet, es, col, row);
      if (v is NumberValue || v is TextValue)
        act(v);
      else if (v is MatrixValue)
        (v as MatrixValue).Apply(act);
      else if (v != null)
        throw new ArgumentException();   // Non-null but wrong type
    }

    public override Expr Move(int deltaCol, int deltaRow) {
      Expr[] newEs = new Expr[es.Length];
      for (int i = 0; i < es.Length; i++)
        newEs[i] = es[i].Move(deltaCol, deltaRow);
      return new FunCall(function, newEs);
    }

    public override Adjusted<Expr> InsertRowCols(Sheet modSheet, bool thisSheet,
                                                 int R, int N, int r,
                                                 bool doRows) 
    {
      Expr[] newEs = new Expr[es.Length];
      int upper = int.MaxValue;
      bool same = true;
      for (int i = 0; i < es.Length; i++) {
        Adjusted<Expr> ae
          = es[i].InsertRowCols(modSheet, thisSheet, R, N, r, doRows);
        upper = Math.Min(upper, ae.upper);
        same = same && ae.same;
        newEs[i] = ae.e;
      }
      return new Adjusted<Expr>(new FunCall(function, newEs), upper, same);
    }

    // Show infixed operators as infix and without excess parentheses

    public override String Show(int col, int row, int ctxpre, Formats fo) {
      StringBuilder sb = new StringBuilder();
      int pre = function.fixity;
      if (pre == 0) { // Not operator
        sb.Append(function.name).Append("(");
        for (int i = 0; i < es.Length; i++) {
          if (i > 0)
            sb.Append(";");
          sb.Append(es[i].Show(col, row, 0, fo));
        }
        sb.Append(")");
      }
      else { // Operator.  Assume es.Length is 1 or 2 
        if (es.Length == 2) {
          // If precedence lower than context, add parens
          if (pre < ctxpre)
            sb.Append("(");
          sb.Append(es[0].Show(col, row, pre, fo));
          sb.Append(function.name);
          // Only higher precedence right operands avoid parentheses
          sb.Append(es[1].Show(col, row, pre + 1, fo));
          if (pre < ctxpre)
            sb.Append(")");
        }
        else if (es.Length == 1) {
          sb.Append(function.name);
          sb.Append(es[1].Show(col, row, pre, fo));
        }
        else
          throw new ImpossibleException("Operator not unary or binary");
      }
      return sb.ToString();
    }
  }

  // ----------------------------------------------------------------
  // A CellRef expression is A1 or $A1 or A$1 or $A$1 or Sheet1!A1

  class CellRef : Expr {
    private readonly RARef raref;
    private readonly Sheet sheet;    // non-null if sheet-absolute

    public CellRef(Sheet sheet, RARef raref) {
      this.sheet = sheet;
      this.raref = raref;
    }

    public CellRef(Sheet sheet, bool colAbs, int colRef, bool rowAbs, int rowRef)
      : this(sheet, new RARef(colAbs, colRef, rowAbs, rowRef)) {
    }

    // Evaluate cell ref by evaluating the cell referred to 

    public override Value Eval(Sheet sheet, int col, int row) {
      if (this.sheet != null)
        sheet = this.sheet;
      CellAddr ca = raref.Addr(col, row);
      Cell cell = sheet[ca];
      return cell == null ? null : cell.Eval(sheet, ca.col, ca.row);
    }

    // Apply to cell ref by applying to the cell referred to 

    public override void Apply(Sheet sheet, int col, int row, Act<Value> act) {
      if (this.sheet != null)
        sheet = this.sheet;
      CellAddr ca = raref.Addr(col, row);
      Cell cell = sheet[ca];
      if (cell != null) 
        cell.Apply(sheet, ca.col, ca.row, act);
    }

    // Clone and move (when the containing formula is moved, not copied!)

    public override Expr Move(int deltaCol, int deltaRow) {
      return new CellRef(sheet, raref.Move(deltaCol, deltaRow));
    }

    public override Adjusted<Expr> InsertRowCols(Sheet modSheet, bool thisSheet,
                                                 int R, int N, int r,
                                                 bool doRows) 
    {
      if (sheet == modSheet || sheet == null && thisSheet) {
        Adjusted<RARef> adj = raref.InsertRowCols(R, N, r, doRows);
        return new Adjusted<Expr>(new CellRef(sheet, adj.e),
                                  adj.upper, adj.same);
      } else
        return new Adjusted<Expr>(this);
    }

    public override String Show(int col, int row, int ctxpre, Formats fo) {
      String s = raref.Show(col, row, fo);
      return sheet == null ? s : sheet.Name + "!" + s;
    }
  }

  // ----------------------------------------------------------------
  // A CellArea is an expression such as 
  // A1:C4 or $A$1:C4 or A1:$C4 or Sheet1!A1:C4

  class CellArea : Expr {
    private readonly RARef ul, lr;  // upper left, lower right
    private readonly Sheet sheet;   // non-null if sheet-absolute

    public CellArea(Sheet sheet,
                    bool ulColAbs, int ulColRef, bool ulRowAbs, int ulRowRef,
                    bool lrColAbs, int lrColRef, bool lrRowAbs, int lrRowRef)
      : this(sheet,
             new RARef(ulColAbs, ulColRef, ulRowAbs, ulRowRef),
             new RARef(lrColAbs, lrColRef, lrRowAbs, lrRowRef)) {
    }

    public CellArea(Sheet sheet, RARef ul, RARef lr) {
      this.sheet = sheet;
      this.ul = ul;
      this.lr = lr;
    }

    // Evaluate cell area by evaluating every cell in the area

    public override Value Eval(Sheet sheet, int col, int row) {
      CellAddr ulCa = ul.Addr(col, row), lrCa = lr.Addr(col, row);
      CellAddr.NormalizeArea(ulCa, lrCa, out ulCa, out lrCa);
      int cols = lrCa.col - ulCa.col + 1, rows = lrCa.row - ulCa.row + 1;
      int col0 = ulCa.col, row0 = ulCa.row;
      Value[,] values = new Value[cols, rows];
      for (int c = 0; c < cols; c++)
        for (int r = 0; r < rows; r++) {
          Cell cell = sheet[col0 + c, row0 + r];
          if (cell != null)
            values[c, r] = cell.Eval(sheet, col0 + c, row0 + r);
        }
      return new MatrixValue(values);
    }

    // Apply to cell area by applying to every cell in the area

    public override void Apply(Sheet sheet, int col, int row, Act<Value> act) {
      CellAddr ulCa = ul.Addr(col, row), lrCa = lr.Addr(col, row);
      CellAddr.NormalizeArea(ulCa, lrCa, out ulCa, out lrCa);
      int cols = lrCa.col - ulCa.col + 1, rows = lrCa.row - ulCa.row + 1;
      int col0 = ulCa.col, row0 = ulCa.row;
      for (int c = 0; c < cols; c++)
        for (int r = 0; r < rows; r++) {
          Cell cell = sheet[col0 + c, row0 + r];
          if (cell != null)
            cell.Apply(sheet, col0 + c, row0 + r, act);
        }
    }

    // Clone and move (when the containing formula is moved, not copied)

    public override Expr Move(int deltaCol, int deltaRow) {
      return new CellArea(sheet,
                          ul.Move(deltaCol, deltaRow),
                          lr.Move(deltaCol, deltaRow));
    }

    public override Adjusted<Expr> InsertRowCols(Sheet modSheet, bool thisSheet,
                                                 int R, int N, int r,
                                                 bool doRows) 
    {
      if (sheet == modSheet || sheet == null && thisSheet) {
        Adjusted<RARef> ulNew = ul.InsertRowCols(R, N, r, doRows),
                        lrNew = lr.InsertRowCols(R, N, r, doRows);
        int upper = Math.Min(ulNew.upper, lrNew.upper);
        return new Adjusted<Expr>(new CellArea(sheet, ulNew.e, lrNew.e),
                                  upper, ulNew.same && lrNew.same);
      }
      else
        return new Adjusted<Expr>(this);
    }

    public override String Show(int col, int row, int ctxpre, Formats fo) {
      String s = ul.Show(col, row, fo) + ":" + lr.Show(col, row, fo);
      return sheet == null ? s : sheet.Name + "!" + s;
    }
  }
}
