// CoreCalc, a spreadsheet core implementation 

// ----------------------------------------------------------------------
// Copyright (c) 2006 Peter Sestoft and Thomas Iversen

// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without
// restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

//  * The above copyright notice and this permission notice shall be
//    included in all copies or substantial portions of the Software.

//  * The software is provided "as is", without warranty of any kind,
//    express or implied, including but not limited to the warranties of
//    merchantability, fitness for a particular purpose and
//    noninfringement.  In no event shall the authors or copyright
//    holders be liable for any claim, damages or other liability,
//    whether in an action of contract, tort or otherwise, arising from,
//    out of or in connection with the software or the use or other
//    dealings in the software.
// ----------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Text;

// Machinery to define built-in functions and operators

namespace CoreCalc {
  class Function {
    public readonly String name;
    public readonly Applier applier;
    public readonly int fixity; // Non-zero: precedence of operator
    private static readonly IDictionary<String, Function> table;
    private static readonly Random rnd = new Random();
    private static readonly long basedate = new DateTime(1899, 12, 30).Ticks;

    // Get a Function by name
    public static Function Get(String name) {
      return table[name.ToUpper()];
    }

    static Function() {
      table = new Dictionary<String, Function>();
      // <fun> : unit -> number
      new Function("NOW",
                   MakeFunction(delegate() {
                     return (DateTime.Now.Ticks - basedate) * 100E-9 / 60 / 60 / 24;
      }));
      new Function("PI",
                   MakeConstant(3.14159265358979));
      new Function("RAND",
                   MakeFunction(rnd.NextDouble));
      // <fun> : number -> number
      new Function("SIN",
          MakeFunction(Math.Sin));
      new Function("COS",
          MakeFunction(Math.Cos));
      new Function("TAN",
          MakeFunction(Math.Tan));
      new Function("LN",
          MakeFunction((Fun<double,double>)Math.Log));
      new Function("LOG10",
          MakeFunction(Math.Log10));
      new Function("EXP",
          MakeFunction(Math.Exp));
      new Function("SQRT",
          MakeFunction(Math.Sqrt));
      // <fun> : number * number -> number
      new Function("CEILING",
          MakeFunction(delegate(double d, double signif) {
            return signif * Math.Ceiling(d / signif);
          }));
      new Function("FLOOR",
          MakeFunction(delegate(double d, double signif) {
            return signif * Math.Floor(d / signif);
          }));
      new Function("ROUND",
          MakeFunction(delegate(double d, double digits) {
            return Math.Round(d, (int)digits);
          }));
      new Function("^", 8,
          MakeFunction(Math.Pow));
      new Function("*", 7,
          MakeFunction(delegate(double x, double y) { return x * y; }));
      new Function("/", 7,
          MakeFunction(delegate(double x, double y) { return x / y; }));
      new Function("+", 6,
          MakeFunction(delegate(double x, double y) { return x + y; }));
      new Function("-", 6,
          MakeFunction(delegate(double x, double y) { return x - y; }));
      new Function("&", 6,
          MakeFunction(delegate(String x, String y) { return x + y; }));
      new Function("<", 5,
          MakePredicate(delegate(double x, double y) { return x < y; }));
      new Function("<=", 5,
          MakePredicate(delegate(double x, double y) { return x <= y; }));
      new Function(">=", 5,
          MakePredicate(delegate(double x, double y) { return x >= y; }));
      new Function(">", 5,
          MakePredicate(delegate(double x, double y) { return x > y; }));
      new Function("==", 4,
          MakePredicate(delegate(double x, double y) { return x == y; }));
      new Function("<>", 4,
          MakePredicate(delegate(double x, double y) { return x != y; }));
      //  SUM : { number, matrix } * -> number
      new Function("SUM",
          MakeFunction(delegate(Value[] vs) {
            double sum = 0.0;
            Apply(vs, delegate(double x) { sum += x; });
            return sum;
          }));
      new Function("FASTSUM",
          delegate(Sheet sheet, Expr[] es, int col, int row) {
            double sum = 0.0;
            foreach (Expr e in es) {
              e.Apply(sheet, col, row,
                      delegate(Value v) {
                        NumberValue number = v as NumberValue;
                        if (number != null)
                          sum += number.value;
                        else
                          throw new ArgumentException();
                      });
            }
            return new NumberValue(sum);
          });
      //  AVG : (number | matrix)*  -> number
      new Function("AVG",
          MakeFunction(delegate(Value[] vs) {
            double sum = 0.0;
            int count = 0;
            Apply(vs, delegate(double x) { sum += x; count++; });
            return sum / count;
          }));
      // TRANSPOSE : matrix -> matrix
      new Function("TRANSPOSE",
          MakeFunction(delegate(Value v0) {
            MatrixValue v0mat = v0 as MatrixValue;
            if (v0mat != null) {
              int cols = v0mat.Rows, rows = v0mat.Cols;
              Value[,] res = new Value[cols, rows];
              for (int c = 0; c < cols; c++)
                for (int r = 0; r < rows; r++)
                  res[c, r] = v0mat[r, c];
              return new MatrixValue(res);
            } else
              return new ErrorValue("ARGTYPE");
          }));
      // COLS : matrix -> number  // Too strict?  If arg is CellArea
      new Function("COLS",
          MakeFunction(delegate(Value v0) {
            MatrixValue v0mat = v0 as MatrixValue;
            if (v0mat != null)
              return new NumberValue(v0mat.Cols);
            else
              return new ErrorValue("ARGTYPE");
          }));
      // ROWS : matrix -> number  // Too strict?  If arg is CellArea
      new Function("ROWS",
          MakeFunction(delegate(Value v0) {
            MatrixValue v0mat = v0 as MatrixValue;
            if (v0mat != null)
              return new NumberValue(v0mat.Rows);
            else
              return new ErrorValue("ARGTYPE");
          }));
      // ISMATRIX : matrix -> number  // Too strict?  If arg is CellArea
      new Function("ISMATRIX",
          MakeFunction(delegate(Value v0) {
            if (v0 != null)
              return new NumberValue(v0 is MatrixValue ? 1.0 : 0.0);
            else
              return new ErrorValue("ARGNULL");
          }));
      // IF : number any any -> any, 
      new Function("IF",                  // Note: non-strict in arg 2 and 3
          delegate(Sheet sheet, Expr[] es, int col, int row) {
            if (es.Length == 3) {
              NumberValue v0 = es[0].Eval(sheet, col, row) as NumberValue;
              if (v0 != null)
                if (v0.value != 0)
                  return es[1].Eval(sheet, col, row);
                else
                  return es[2].Eval(sheet, col, row);
              else
                return new ErrorValue("ARGTYPE");
            }
            else
              return new ErrorValue("ARGCOUNT");
          });
      // CHOOSE number * any* -> any
      new Function("CHOOSE",              // Note: non-strict in arg 2...
          delegate(Sheet sheet, Expr[] es, int col, int row) {
            if (es.Length >= 1) {
              NumberValue v0 = es[0].Eval(sheet, col, row) as NumberValue;
              if (v0 != null && 1 <= v0.value && v0.value < es.Length)
                return es[(int)(v0.value)].Eval(sheet, col, row);
              else
                return new ErrorValue("ARGTYPE");
            }
            else
              return new ErrorValue("ARGCOUNT");
          });
    }
    private Function(String name, Applier applier)
      : this(name, 0, applier) { }

    private Function(String name, int fixity, Applier applier) {
      this.name = name;
      this.applier = applier;
      this.fixity = fixity;
      table.Add(name, this);
    }

    // Make number-valued nullary function, eg RAND()
    private static Applier MakeFunction(Fun<double> dlg) {
      return
        delegate(Sheet sheet, Expr[] es, int col, int row) {
          if (es.Length == 0)
            return new NumberValue(dlg());
          else
            return new ErrorValue("ARGCOUNT");
        };
    }

    // Make number-valued constant nullary function, eg PI()
    private static Applier MakeConstant(double d) {
      NumberValue value = new NumberValue(d);
      return
        delegate(Sheet sheet, Expr[] es, int col, int row) {
          if (es.Length == 0)
            return value;
          else
            return new ErrorValue("ARGCOUNT");
        };
    }

    // Make number-valued strict unary function, eg SIN(x)
    private static Applier MakeFunction(Fun<double, double> dlg) {
      return
        delegate(Sheet sheet, Expr[] es, int col, int row) {
          if (es.Length == 1) {
            NumberValue v0 = es[0].Eval(sheet, col, row) as NumberValue;
            if (v0 != null)
              return new NumberValue(dlg(v0.value));
            else
              return new ErrorValue("ARGTYPE");
          }
          else
            return new ErrorValue("ARGCOUNT");
        };
    }

    // Make number-valued strict binary function, eg +
    private static Applier MakeFunction(Fun<double, double, double> dlg) {
      return
        delegate(Sheet sheet, Expr[] es, int col, int row) {
          if (es.Length == 2) {
            NumberValue v0 = es[0].Eval(sheet, col, row) as NumberValue,
                        v1 = es[1].Eval(sheet, col, row) as NumberValue;
            if (v0 != null && v1 != null)
              return new NumberValue(dlg(v0.value, v1.value));
            else
              return new ErrorValue("ARGTYPE");
          }
          else
            return new ErrorValue("ARGCOUNT");
        };
    }

    // Make string-valued strict binary function, eg &
    private static Applier MakeFunction(Fun<String, String, String> dlg) {
      return
        delegate(Sheet sheet, Expr[] es, int col, int row) {
          if (es.Length == 2) {
            TextValue v0 = es[0].Eval(sheet, col, row) as TextValue,
                      v1 = es[1].Eval(sheet, col, row) as TextValue;
            if (v0 != null && v1 != null)
              return new TextValue(dlg(v0.value, v1.value));
            else
              return new ErrorValue("ARGTYPE");
          }
          else
            return new ErrorValue("ARGCOUNT");
        };
    }

    // Make boolean-valued strict binary function, eg ==
    private static Applier MakePredicate(Fun<double, double, bool> dlg) {
      return
        MakeFunction(delegate(double x, double y) {
          return dlg(x, y) ? 1.0 : 0.0;
        });
    }

    // Make number-valued strict variadic function, eg SUM(A1; B2:B7)
    private static Applier MakeFunction(Fun<Value[], double> dlg) {
      return
        delegate(Sheet sheet, Expr[] es, int col, int row) {
          try {
            return new NumberValue(dlg(Eval(es, sheet, col, row)));
          } catch (ArgumentException) {
            return new ErrorValue("ARGTYPE");
          }
        };
    }

    // Make number-valued strict monadic function, eg ROWS(A1:B7)
    private static Applier MakeFunction(Fun<Value, Value> dlg) {
      return
        delegate(Sheet sheet, Expr[] es, int col, int row) {
          if (es.Length == 1)
            return dlg(es[0].Eval(sheet, col, row));
          else
            return new ErrorValue("ARGCOUNT");
        };
    }

    // Make general strict function, eg MMULT(A1:B3, E1:G2)
    private static Applier MakeFunction(Fun<Value[], Value> dlg) {
      return
        delegate(Sheet sheet, Expr[] es, int col, int row) {
          return dlg(Eval(es, sheet, col, row));
        };
    }

    // Evaluate expression array
    private static Value[] Eval(Expr[] es, Sheet sheet, int col, int row) {
      Value[] vs = new Value[es.Length];
      for (int i = 0; i < es.Length; i++)
        vs[i] = es[i].Eval(sheet, col, row);
      return vs;
    }

    // Apply act to numbers in vs, and recursively to matrix values
    private static void Apply(Value[] vs, Act<double> act) {
      foreach (Value v in vs) {
        if (v is NumberValue)
          act((v as NumberValue).value);
        else if (v is MatrixValue)
          (v as MatrixValue).Apply(act);
        else
          throw new ArgumentException();
      }
    }
  }
}
