// CoreCalc, a spreadsheet core implementation 

// ----------------------------------------------------------------------
// Copyright (c) 2006 Peter Sestoft

// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without
// restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

//  * The above copyright notice and this permission notice shall be
//    included in all copies or substantial portions of the Software.

//  * The software is provided "as is", without warranty of any kind,
//    express or implied, including but not limited to the warranties of
//    merchantability, fitness for a particular purpose and
//    noninfringement.  In no event shall the authors or copyright
//    holders be liable for any claim, damages or other liability,
//    whether in an action of contract, tort or otherwise, arising from,
//    out of or in connection with the software or the use or other
//    dealings in the software.
// ----------------------------------------------------------------------

using System;
using System.Text;

// Class Value and its subclasses are used to represent the possible 
// values of a spreadsheet cell

namespace CoreCalc {

  // A Value is the result of evaluating an expression

  public abstract class Value { }

  // ----------------------------------------------------------------
  // A NumberValue is a floating-point number

  class NumberValue : Value {
    public readonly double value;

    public NumberValue(double value) {
      this.value = value;
    }

    public override String ToString() {
      return value.ToString();
    }
  }

  // ----------------------------------------------------------------
  // A TextValue is a string

  class TextValue : Value {
    public readonly String value;

    public TextValue(String value) {
      this.value = value;
    }

    public override String ToString() {
      return value;
    }
  }

  // ----------------------------------------------------------------
  // An ErrorValue is a value indicating failure of evaluation

  class ErrorValue : Value {
    private readonly String msg;

    public ErrorValue(String msg) {
      this.msg = "#ERR: " + msg;
    }

    public override String ToString() {
      return msg;
    }
  }

  // ----------------------------------------------------------------
  // A MatrixValue is a rectangle of Values 

  class MatrixValue : Value {
    private readonly Value[,] values;   // Non-null; elements may be null

    public MatrixValue(Value[,] values) {
      this.values = values;
    }

    public int Cols {
      get { return values.GetLength(0); }
    }

    public int Rows {
      get { return values.GetLength(1); }
    }

    // Get cell value from matrix value
    public Value this[CellAddr ca] {
      get {
        //    0 <= ca.col && ca.col < Cols
        // && 0 <= ca.row && ca.row < Rows
        return values[ca.col, ca.row];
      }
    }

    public Value this[int col, int row] {
      get {
        //    0 <= col && col < Cols
        // && 0 <= row && row < Rows
        return values[col, row];
      }
    }

    public void Apply(Act<double> act) {
      foreach (Value v in values)
        if (v != null) // Only active cells contribute
          if (v is NumberValue)
            act((v as NumberValue).value);
          else if (v is MatrixValue)
            (v as MatrixValue).Apply(act);
          else if (v is ErrorValue)
              throw new ErrorValueException(v as ErrorValue);
          else
            throw new Exception("MatrixValue.Apply");
    }

    public void Apply(Act<Value> act) {
      foreach (Value v in values)
        if (v != null) // Only active cells contribute
          if (v is NumberValue || v is TextValue)
            act(v);
          else if (v is MatrixValue)
            (v as MatrixValue).Apply(act);
          else if (v is ErrorValue)
            throw new ErrorValueException(v as ErrorValue);
          else
            throw new Exception("MatrixValue.Apply");
    }

    public override String ToString() {
      StringBuilder sb = new StringBuilder();
      for (int c = 0; c < Cols; c++) {
        for (int r = 0; r < Rows; r++) {
          Value v = values[c, r];
          sb.Append(v == null ? "[none]" : v.ToString());
        }
        sb.Append("\n");
      }
      return sb.ToString();
    }
  }
}
